
#include <ESP8266WiFi.h>
#include <OneWire.h>

#include <brzo_i2c.h>
#include "SSD1306Brzo.h"

#include <Adafruit_NeoPixel.h>

#include <SoftwareSerial.h>

extern "C" {
  #include "user_interface.h"
  #include "config.h"
}

#define FPM_SLEEP_MAX_TIME 0xFFFFFFF

#define M_ESP_SETUP 0
#define M_ESP_RUN   1

#define SDA_PIN D2
#define SDC_PIN D5
#define DS_PIN D6

#define MHZ19_RX_PIN D7
#define MHZ19_TX_PIN D8

#define WS2812B_PIN D3
#define NUMPIXELS   1

// Display Settings
// OLED will be checked with this address and this address+1
// so here 0x03c and 0x03d
#define I2C_DISPLAY_ADDRESS 0x3c

SSD1306Brzo display(I2C_DISPLAY_ADDRESS, SDA_PIN, SDC_PIN);
SoftwareSerial co2serial( MHZ19_RX_PIN, MHZ19_TX_PIN );
Adafruit_NeoPixel pixels = Adafruit_NeoPixel(NUMPIXELS, WS2812B_PIN, NEO_GRB + NEO_KHZ800);
OneWire  ds(DS_PIN);
static int save_ds_eeprom=1;
static uint32_t ds_started=0;

#define RGB(r,g,b)  (((uint32_t)(r)<<16)|((uint32_t)(g)<<8)|(b))

#define DGREEN   RGB(0,255,0)
#define DRED     RGB(255,0,0)
#define DYELLOW  RGB(255,255,0)
#define DBLUE    RGB(0,0,255)
#define DBLACK   0

byte addr[8];
byte ds_scan=1;

/* 12bit need 750 */
#define SAMPLE_TIME   1000

void startGetTempDs18b20( void )
{
  float temp;
  int   t100;
  int   i;
#if 0
//  byte addr[8] = { 0x28, 0xff, 0x68, 0xde, 0xb5, 0x16, 0x03, 0x5e };
  byte addr[8] = { 0x28, 0xff, 0x95, 0xd1, 0xb5, 0x16, 0x03, 0x8e };
#endif

  if ( ds_scan )
  {
    ds.search(addr);
    if (OneWire::crc8(addr, 7) != addr[7])
    {
#ifdef SERIAL_DEBUG
       Serial.println("CRC is not valid!");
#endif
       return;
    }
	ds_scan=0;
#ifdef SERIAL_DEBUG
    Serial.print("ds-search : ");
    for(i=0;i<8;i++)
      Serial.printf(" %02x ",addr[i]);
    Serial.println("");
#endif
  }
  ds.reset();

  if ( save_ds_eeprom )
  {   /* save resolution of 9bit into eeprom of ds18b20 */
    save_ds_eeprom=0;

    ds.select(addr);
    ds.write(0x4E);  // set config
    ds.write(0x00);  // unused
    ds.write(0x00);  // unused
    ds.write(0x7F);  // 1f=9bit, 3f=10bit, 5f=11bit, 7f=12bit
                     // 93.75ms   187.5ms    375ms      750ms
    ds.reset();
    ds.select(addr);
    ds.write(0x48);  // save into eeprom
    ds.reset();
    delay(15);
  }

  ds.select(addr);
  ds.write(0x44, 1);

  ds_started=millis() | 0x80000000;
}

int16_t getTempDs18b20( void )
{
  uint32_t  now=millis() | 0x80000000;
  int       i;
  int16_t   t10=0;
#if 0
//  byte    addr[8] = { 0x28, 0xff, 0x68, 0xde, 0xb5, 0x16, 0x03, 0x5e };
  byte      addr[8] = { 0x28, 0xff, 0x95, 0xd1, 0xb5, 0x16, 0x03, 0x8e };
#endif
  byte      data[12];
  byte      crc;

  if ( now < ds_started )
  {
     ds_started &= 0x7fffffff;    /* remove bit31 */
     ds_started |= 1;             /* make != 0 */
  }
  if ( now < ds_started + SAMPLE_TIME )
  {
    delay(SAMPLE_TIME+ds_started-now);
  }

  if ( ds_scan )
    return 0;

  ds.reset();
  ds.select(addr);
  ds.write(0xBE);

  for ( i = 0; i < 9; i++)
  {           // we need 9 bytes
     data[i] = ds.read();

#ifdef SERIAL_DEBUG
//Serial.print(data[i], HEX);
//Serial.print(" ");
#endif
  }
  crc=OneWire::crc8(data, 8);

#ifdef SERIAL_DEBUG
//Serial.print(" CRC=");
//Serial.print(crc, HEX);
//Serial.println();
#endif

  if ( crc != data[8] )
    return 0;

  int16_t raw = (data[1] << 8) | data[0];
  byte cfg = (data[4] & 0x60);

#ifdef SERIAL_DEBUG
//Serial.print(" CFG=");
//Serial.println(cfg,HEX);
#endif

  // at lower res, the low bits are undefined, so let's zero them
  if (cfg == 0x00) raw = raw & ~7;  // 9 bit resolution, 93.75 ms
  else if (cfg == 0x20) raw = raw & ~3; // 10 bit res, 187.5 ms
  else if (cfg == 0x40) raw = raw & ~1; // 11 bit res, 375 ms
//// default is 12 bit resolution, 750 ms conversion time
  raw*=10;
  t10=raw>>4;

  return t10;
}

static int      ppm=0;
static int      co2_sample_cnt=0;
static uint32_t cali_started=0;
static int      co2_next_cmd=0;    /* 0=readPPM, 1=calZero, 2=disABC, 3=enaABC */

static uint16_t l10temp[10];

void dsLoop( void )
{
  if ( ds_started )
  {
    uint32_t  now=millis() | 0x80000000;
    uint32_t  diff;

    if ( now < ds_started )
    {
      ds_started &= 0x7fffffff;    /* remove bit31 */
      ds_started |= 1;             /* make != 0 */
    }
    diff=now-ds_started;
    int perc=diff*100/SAMPLE_TIME;
    if ( diff >= SAMPLE_TIME )
    {
      uint16_t dstemp=getTempDs18b20();
      uint16_t temp = dstemp;
      String tstrg;

      display.clear();

/* compensate electronic heating */
      if ( temp > 256 )
        temp -= 37;
      else if ( temp > 182 )
        temp -= (temp - 182)/2;

      memcpy(l10temp,l10temp+1,sizeof(uint16_t)*9);
      l10temp[9] = dstemp;
      if ( ! l10temp[0] )
      {
        int i;
        for( i=0; i<9; i++ )
           l10temp[i] = dstemp;
      }

      if (( dstemp > l10temp[0] ) &&
          ( dstemp - l10temp[0] > 10 ) && ( co2_sample_cnt < 11) )
      {
        cali_started = now | 0x80000000;   /* start calibration of sensor */
        pixels.setPixelColor(0,DBLUE);
        pixels.show();
      }

      tstrg=(temp/10);
      tstrg+=".";
      tstrg+=(temp%10);
      tstrg+="°C";

#ifdef SERIAL_DEBUG
      Serial.printf("Temp: %d , delta=%d\n",dstemp,dstemp - l10temp[0]);
#endif
      
      if ( cali_started )
      {
        uint32_t left;
        if ( now < cali_started )
        {
          cali_started &= 0x7fffffff;
          cali_started |=1;
        }
        diff = (now-cali_started)/1000;
        left =  diff > 1200 ? 0:1200-diff;
        cali_started |= 0x80000000;
        tstrg = "C: ";
        tstrg += left;
        tstrg +="s";
        if ( !left )
        {
          cali_started = 0;
          co2_next_cmd = 1; /* calZero */
        }
        display.setFont(ArialMT_Plain_16);
      }
      else
        display.setFont(ArialMT_Plain_24);

      display.drawString(64,24,tstrg);

      if ( co2_sample_cnt < 10 )
      {
        display.setFont(ArialMT_Plain_16);
        display.drawProgressBar(0,50,120,10,co2_sample_cnt*100/9);

      }
      else
      {
        tstrg=ppm;
        tstrg+="ppm";
        display.drawString(64,50,tstrg);
      }
      display.drawLine(0,36,127,36);

      display.display();

      startGetTempDs18b20();
    }
  }
}

static uint32_t last_co2_stamp=0;

void co2Loop( void )
{
  uint32_t  now=millis() | 0x80000000;
  uint32_t  diff;

  if ( now < last_co2_stamp )
  {
    last_co2_stamp &= 0x7fffffff;    /* remove bit31 */
    last_co2_stamp |= 1;             /* make != 0 */
  }
  diff=now-last_co2_stamp;
	last_co2_stamp |= 0x80000000;

  if ( diff >= CO2_CYCLE )
  {
    byte readPPM[9] = { 0xff, 0x01, 0x86, 0x00,0x00,0x00,0x00,0x00, 0x79 };
    byte calZero[9] = { 0xff, 0x01, 0x87, 0x00,0x00,0x00,0x00,0x00, 0x78 };
    byte enaABC[9] = { 0xff, 0x01, 0x79, 0xA0, 0x00,0x00,0x00,0x00, 0xE6 };
    byte disABC[9] = { 0xff, 0x01, 0x79, 0x00,0x00,0x00,0x00,0x00, 0x86 };
    byte *cmd = disABC;

    if ( co2_next_cmd == 3 ) /* cur.not used */
      cmd = enaABC;
    else if ( co2_next_cmd == 2 )
      cmd = disABC;
    else if ( co2_next_cmd == 1 )
      cmd = calZero;
    else
      cmd = readPPM;

    co2_next_cmd = 0;

    last_co2_stamp = now;
    co2serial.write(cmd,9);
  }

  if ( !co2serial.available() )
     return;
  while (co2serial.available() > 0 &&
       (unsigned char)co2serial.peek() != 0xFF)
  {
     co2serial.read();
  }
  byte response[9];
  int  i;

  memset(response,0,9);
  co2serial.readBytes(response,9);
#ifdef SERIAL_DEBUG
  Serial.print("[" + String(millis()/1000)+"] ");
  Serial.print("co2-data : ");
  for( i=0; i<9; i++ )
    Serial.printf("%02x ",response[i]);
  Serial.println("");  
#endif

  if ( response[1] != 0x86 )
    return;

  byte crc = 0;

  for( i=1; i<8; i++ )
    crc += response[i];
  crc = 255 - crc+1;

  if ( crc == response[8] )
  {
    int temp = response[4] - 40;

    ppm = (int)256*response[2] + response[3];
#ifdef SERIAL_DEBUG
    Serial.printf("PPM: %d, Temp: %d°C\n",ppm,temp);
#endif

    co2_sample_cnt++;
    if ( co2_sample_cnt == 8 )
    {
      co2_next_cmd = 2;   /* next will disable ABC */
    }
    if ( cali_started )
      pixels.setPixelColor(0,DBLUE);
    else if ( co2_sample_cnt > 9 )
    {
      uint32_t col=DGREEN;
      if ( ppm > 999 )
        col = DRED;
      else if ( ppm > 799 )
        col = DYELLOW;
      pixels.setPixelColor(0,col);
    }
    else
      pixels.setPixelColor(0,DBLACK);

    pixels.show();
  }
}

void setup()
{
  int i;

#ifdef SERIAL_DEBUG
  Serial.begin(115200);
#endif

  pixels.begin();
  pixels.setPixelColor(0,DBLACK);
  pixels.show();

  display.init();
  display.flipScreenVertically();
  display.setContrast(255);

  display.clear();
  display.setFont(ArialMT_Plain_24);
  display.setTextAlignment(TEXT_ALIGN_CENTER_BOTH);

  display.drawString(64,22,"Starting...");
  display.setFont(ArialMT_Plain_16);
  display.drawString(64,50,"Co2-Puk : " PVERSION);
  display.display();

  memset(l10temp,0,sizeof(int)*10);

  startGetTempDs18b20();
  co2serial.begin(9600);

  wifi_station_disconnect();
  wifi_set_opmode(NULL_MODE);
  wifi_set_sleep_type(MODEM_SLEEP_T);
  wifi_fpm_open();
  wifi_fpm_do_sleep(FPM_SLEEP_MAX_TIME);
  
  display.clear();

#ifdef SERIAL_DEBUG
  Serial.print(F("FlashSizebyChipID: "));
  Serial.println(ESP.getFlashChipSizeByChipId());
#endif
}

void loop()
{
static int actloops=0;

  dsLoop();
  co2Loop();

  delay(10);
}
